-- make sure new joiners have the hostmod table
-- players.iterate doesn't run for players without an mobj,
-- and this hook runs immediately after a player spawns for the first time
addHook("PlayerSpawn", function(p)
	p.hostmod = $ or {}
end)

-- ...and set it for the dedicated server player so they can start votes
-- not that their vote counts anyway because they don't get iterated over by players.iterate
addHook("MapChange", function(p)
	if isdedicatedserver then server.hostmod = $ or {} end
end)

rawset(_G, "HOSTMOD", "15")
print("HOSTMOD v"..HOSTMOD..", by Tyron. hyuu.cc/contact")

rawset(_G, "hm_escape", {
	{a = "\\q", b = "\""},
	{a = "\\n", b = "\n"},
	{a = "\\128", b = "\128"},
	{a = "\\129", b = "\129"},
	{a = "\\130", b = "\130"},
	{a = "\\131", b = "\131"},
	{a = "\\132", b = "\132"},
	{a = "\\133", b = "\133"},
	{a = "\\134", b = "\134"},
	{a = "\\135", b = "\135"},
	{a = "\\136", b = "\136"},
	{a = "\\137", b = "\137"},
	{a = "\\138", b = "\138"},
	{a = "\\139", b = "\139"},
	{a = "\\140", b = "\140"},
	{a = "\\141", b = "\141"},
	{a = "\\142", b = "\142"},
	{a = "\\143", b = "\143"},
})

-- From LJ Sonik's PRNG library
rawset(_G, "hm_prng_t", 0)
rawset(_G, "hm_prng_x", 6278)
rawset(_G, "hm_prng_y", 975)
rawset(_G, "hm_prng_z", 39207)
rawset(_G, "hm_prng_w", 45678)
rawset(_G, "HM_Random", function()
	hm_prng_t = hm_prng_x ^^ (hm_prng_x << 11)
	hm_prng_x, hm_prng_y, hm_prng_z, hm_prng_w = hm_prng_y, hm_prng_z, hm_prng_w, hm_prng_w ^^ (hm_prng_w >> 19) ^^ hm_prng_t ^^ (hm_prng_t >> 8) -- What the fuck?
	return hm_prng_w
end)
rawset(_G, "HM_RandomRange", function(a, b)
    return a + abs(HM_Random() % (b - a + 1))
end)

local pinggfx
local function drawPing(v, x, y, scale, num, flags)
    if pinggfx == nil then
        pinggfx = {}
        
        for i = 1, 5 do
            table.insert(pinggfx, v.cachePatch("HMPING"..i))
        end
    end

    v.drawScaled(x<<FRACBITS, y<<FRACBITS, scale, pinggfx[num], flags)
end

local function haveItemLogs(p)
	if not (p.hostmod and p.hostmod.itemlog and next(p.hostmod.itemlog)) then return false end

	return true
end

local function isHpmodEliminated(p)
    if not (hpmod and hpmod.running) then return false end
    if p.spectator then return false end
    if p.sinked then return false end
    if not (p.pflags & PF_TIMEOVER) then return false end
    if not (p.hpmod and p.hpmod.deathtimer > 0) then return false end
    
    return true
end

local function isNoContest(p)
    if not (p.pflags & PF_TIMEOVER) then return false end
    if p.sinked then return false end
    if isHpmodEliminated(p) then return false end
    
    return true
end

local function getPlayerRanks(spectators)
    local playerranks = {}

    -- Add players that are alive
    for plr in players.iterate do
        if not (plr.spectator or plr.pflags & PF_TIMEOVER) then
			table.insert(playerranks, plr)
        end
    end

    -- sort 'em
    table.sort(playerranks, function(a, b) return a.kartstuff[k_position] < b.kartstuff[k_position] end)

    -- Add players eliminated in elimination gamemode
    local elimplayers = {}
    for plr in players.iterate do
        if Elimination and Elimination() and plr.elim.iseliminated then
            table.insert(elimplayers, plr)
        end
    end

    -- (eliminatedtime or 0) - juuuust in case player has nil in that field.
    -- Dunno if thats possible but good to have protection just in case
    table.sort(elimplayers, function(a, b) return (a.elim.eliminatedtime or 0) > (b.elim.eliminatedtime or 0) end)

    for i, player in ipairs(elimplayers) do
        table.insert(playerranks, player)
    end

    -- Add players eliminated by kitchen sink
    for plr in players.iterate do
        if not plr.spectator and plr.sinked then
            table.insert(playerranks, plr)
        end
    end

    -- Add players eliminated by hp
    for plr in players.iterate do
        if not plr.spectator and isHpmodEliminated(plr) then
            table.insert(playerranks, plr)
        end
    end

    -- Add NO CONTEST'd players
    for plr in players.iterate do
        if not plr.spectator and isNoContest(plr) then
            table.insert(playerranks, plr)
        end
    end

    -- Add the rest of spectators
    for plr in players.iterate do
        if plr.spectator and not (Elimination and Elimination() and plr.elim.iseliminated) and (spectators or haveItemLogs(plr)) then
            table.insert(playerranks, plr)
        end
    end

    return playerranks
end

local function getTimeoverText(p)
	if p.hostmod and p.hostmod.timeovertext then
		return p.hostmod.timeovertext
    elseif p.sinked then
        return "\133SINKED"
    elseif isHpmodEliminated(p) then
        return "\133ELIMINATED"
    end
    
    return "NO CONTEST"
end

local function colorstr(...)
	if not ... then
		return ""
	end

    local msg_buff = {}

    for _, arg in ipairs {...} do
        local arg_buff = {}

        local pos = 0
        local len = #arg

        while pos <= len do
            local char = string.sub(arg, pos, pos)

            if char == '\\' then
                char = string.char(tonumber(string.sub(arg, pos+1, pos+3)))

                pos = pos + 3
            end

            pos = pos + 1

            table.insert(arg_buff, char)
        end

        table.insert(msg_buff, table.concat(arg_buff, ""))
    end

	return table.concat(msg_buff, " ")
end

rawset(_G, "hm_shared", {
    drawPing = drawPing,
    haveItemLogs = haveItemLogs,
    isHpmodEliminated = isHpmodEliminated,
    isNoContest = isNoContest,
    getPlayerRanks = getPlayerRanks,
    getTimeoverText = getTimeoverText,
	colorstr = colorstr,
})
