local function doEscapeCode(str)
	for _, v in pairs(hm_escape)
		str = str:gsub(v.a, v.b)
	end
	return str
end

local hostmod_motd = CV_RegisterVar({
	name = "hm_motd",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})
local hostmod_motdalways = CV_RegisterVar({
	name = "hm_motd_nag",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})
local hostmod_motdname = CV_RegisterVar({
	name = "hm_motd_name",
	defaultvalue = "\136Default HOSTMOD MOTD",
	flags = CV_NETVAR|CV_CALL,
	possiblevalue = nil,
	func = function(var) if (var.changed) and (server) then server.HMNMname = doEscapeCode(var.string) end end
})
local hostmod_motdcontact = CV_RegisterVar({
	name = "hm_motd_contact",
	defaultvalue = "Check your chat?",
	flags = CV_NETVAR|CV_CALL,
	possiblevalue = nil,
	func = function(var) if (var.changed) and (server) then server.HMNMcontact = doEscapeCode(var.string) end end
})
local hostmod_motddesc = CV_RegisterVar({
	name = "hm_motd_tagline",
	defaultvalue = "HOSTMOD's MOTD feature hasn't been configured. Let the host know!",
	flags = CV_NETVAR|CV_CALL,
	possiblevalue = nil,
	func = function(var) if (var.changed) and (server) then server.HMNMdesc = doEscapeCode(var.string) end end
})
local hostmod_motdbackground = CV_RegisterVar({
	name = "hm_motd_bg",
	defaultvalue = "Black",
	flags = CV_NETVAR,
	possiblevalue = SG_CVSkinColors
})

local function doEscapeCodes()
	if not server then return end
	server.HMNMname = doEscapeCode(hostmod_motdname.string)
	server.HMNMcontact = "\134Contact: "..doEscapeCode(hostmod_motdcontact.string)
	server.HMNMdesc = doEscapeCode(hostmod_motddesc.string)
end

doEscapeCodes()

addHook("MapChange", do
	doEscapeCodes()
	if not hostmod_motdalways.value then return end
	for p in players.iterate
		p.HMNMtime = 0
	end
end)

addHook("ThinkFrame", do
	if server.HMNMname == nil then -- WHY?????
		doEscapeCodes()
	end
	for p in players.iterate do
		if not p.HMNMforced and p.spectator then
			COM_BufAddText(p, "view #01") -- make sure freecam control guide doesn't overlap motd
		end
		p.HMNMforced = true
		p.HMNMtime = $ or 0
		p.HMNMtime = $+1
	end
end)

local delay = 4*TICRATE
local maxtime = 12*TICRATE
local scrolldelay = 3*TICRATE

local function xoff(i, o)
	local c = i-o
	if (i < o) or (i > maxtime) return 9999 end
	local x = 0
	if i > maxtime-10 then
		x = maxtime-i-10
	elseif c < 10 then
		x = 10-c
	end
	return x
end

local function aoff(i, o)
	if i < o return V_10TRANS end
	local c = i-o
	if i > maxtime - 10 then
		return (9 - (maxtime - i)) << V_ALPHASHIFT
	end
	if c > 9 return 0 end
	return (9-c) << V_ALPHASHIFT
end

hud.add(function(v, p, c)
	if not hostmod_motd.value then return end
	if not consoleplayer.valid then return end
	local t = consoleplayer.HMNMtime
	if not t then return end
	if t < delay then return end
	t = t - delay
	if t > maxtime+35 then return end
	local bg = v.cachePatch("MOTDBG")
	v.drawScaled((xoff(t,0))*FRACUNIT, 179*FRACUNIT, FRACUNIT, bg, V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_30TRANS|aoff(t,0), v.getColormap(TC_DEFAULT, hostmod_motdbackground.value))
	v.drawString(3+xoff(t, 0), 182, server.HMNMname, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT|aoff(t, 0))
	v.drawString(297+xoff(t, 25), 182, server.HMNMcontact, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTORIGHT|aoff(t, 25), "thin-right")
	
	-- why the fuck did i implement this
	local w = v.stringWidth(server.HMNMdesc, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT|aoff(t, 0), "small")
	local scrolloff = 0
	local realwidth = v.width() / v.dupx()
	if w > realwidth and t > scrolldelay then
		local toscroll = realwidth - w - 25 -- overscroll fudge factor
		local scrollwindow = maxtime - 2*scrolldelay
		local scrollpos = min(t - scrolldelay, maxtime - scrolldelay)
		local scrollpercent = (10000*scrollpos) / (100*scrollwindow)
		scrolloff = toscroll * scrollpercent / 100
	end
	v.drawString(3+xoff(t, 0)+scrolloff, 193, server.HMNMdesc, V_ALLOWLOWERCASE|V_SNAPTOBOTTOM|V_SNAPTOLEFT|aoff(t, 0), "small")
end)