local starttime = 6 * TICRATE + (3 * TICRATE / 4)
local boringskins = {["sonic"] = 1, ["tails"] = 1, ["knuckles"] = 1, ["eggman"] = 1, ["metalsonic"] = 1, ["flicky"] = 1, ["motobug"] = 1, ["amy"] = 1, ["mighty"] = 1, ["ray"] = 1, ["espio"] = 1, ["vector"] = 1, ["chao"] = 1, ["gamma"] = 1, ["chaos"] = 1, ["shadow"] = 1, ["rouge"] = 1, ["herochao"] = 1, ["darkchao"] = 1, ["cream"] = 1, ["omega"] = 1, ["blaze"] = 1, ["silver"] = 1, ["wonderboy"] = 1, ["arle"] = 1, ["nights"] = 1, ["sakura"] = 1, ["ulala"] = 1, ["beat"] = 1, ["vyse"] = 1, ["aiai"] = 1, ["kiryu"] = 1, ["aigis"] = 1, ["miku"] = 1, ["doom"] = 1}
local IRONMAN_OFF, IRONMAN_ON, IRONMAN_JART = 0, 1, 2
local default_randomskins_file = "randomcharacters.txt"

-- If randomcharacters.txt doesn't exist, make it so user don't have to
local function makeRandomSkinsFile()
	local testfile = io.open(default_randomskins_file, 'r')
	
	-- File exists
	if testfile then
		testfile:close()
		return
	end

	local file = assert(io.open(default_randomskins_file, 'w'))
	
	file:write("// Put a list of the characters you want to use below seprated by new lines")
	file:write("// To define a color put a space after the skin on the same line. Example: tails blue")
	file:write("// No color means it will use your current skin color. Use default as a color for the skin's default color.")

	file:close()
end

-- We don't want to terminate our script at startup do we
local ok, err = pcall(makeRandomSkinsFile)
if not ok then
	print("\130WARNING:\128 randomskin.lua: failed to create "..default_randomskins_file..": "..err)
end

-- Remove trailing and leading spaces
local function trim(s)
	return s:match("^%s*(.-)%s*$")
end

local function readRandomSkins(filename)
	local file, err = assert(io.open(filename, 'r'))
	
	local result = {}
	
	for line in file:lines() do
		local comment_start = line:find("//")
		
		-- Skip/remove comments
		if comment_start == 1 then continue end
		if comment_start then line = line:sub(1, comment_start-1) end
		
		line = trim(line)
		
		-- Empty line
		-- "  // Something like this"
		if #line == 0 then continue end
		
		local skin_end = line:find(" ")
		
		// Entire line is just skin end
		if not skin_end then skin_end = #line+1 end
		
		local skin = line:sub(1, skin_end-1)
		line = line:sub(skin_end, #line)
		
		local color
		
		-- We also have color!
		if #line ~= 0 then
			-- Normalize, just in case
			color = (line:lower() == "default") and "default" or line
		end
		
		table.insert(result, {
			skin = trim(skin),
			color = color and trim(color),
		})
	end
	
	file:close()
	
	return result
end

local function ironreminder(p)
	if not p.HMrandomcount then p.HMrandomcount = 0 end
	p.HMrandomcount = $ + 1
	if p.HMrandomcount == 10 then
		S_StartSound(nil, sfx_cdfm20, p)
		CONS_Printf(p, "Why not try \130ironman \128in console to automatically randomize every round?")
	end
end

local function randomskin(p, file, force)
	local newskin

	if not file then
		local index = HM_RandomRange(1, #skins) - 1
		newskin = skins[index]
		ironreminder(p)
	elseif file == "help" then
		CONS_Printf(p, "Usage:")
		CONS_Printf(p, "randomskin - change your skin to a random one")
		CONS_Printf(p, "randomskin <filename> - load list of skins from filename")
		CONS_Printf(p, "randomskin file - shortcut for \"randomskin randomcharacters.txt\"")
		return
	else
		-- `randomskin file` is a shortcut for `randomskin randomcharacters.txt`
		-- Game wouldn't allow to open a file called "file" anyway so why not
		file = file:lower() == "file" and "randomcharacters.txt" or file
		
		-- Technically, we don't have to reload file every time but this allows user to modify
		-- file and have changes applied without having to restart their client
		local ok, result = pcall(readRandomSkins, file)
		
		if not ok then
			CONS_Printf(p, "Error parsing "..file..": "..tostring(result))
			return
		end
		
		if #result == 0 then
			CONS_Printf(p, "File "..file.." has no skins")
			return
		end
		
		local entry = result[HM_RandomRange(1, #result)]
		
		newskin = skins[entry.skin]
		
		if newskin == nil then
			CONS_Printf(p, "Skin "..entry.skin.." does not exist")
			return
		end
		
		if entry.color then
			local newcolor = entry.color
			
			if entry.color == "default" then
				newcolor = tostring(newskin.prefcolor)
			end
			
			COM_BufAddText(p, "color "..newcolor)
		end
	end
	
	S_StartSound(nil, newskin.soundsid[sfxinfo[sfx_khitem].skinsound], p)
	local sanitizedcolor = SG_Color2Chat[newskin.prefcolor] or ""
	COM_BufAddText(p, "skin "..newskin.name)
	CONS_Printf(p, "Trying to change to "..sanitizedcolor..newskin.realname.." \128("..newskin.kartspeed.."/"..newskin.kartweight..")...")
end
COM_AddCommand("randomskin", randomskin, COM_LOCAL)
COM_AddCommand("randomchar", randomskin, COM_LOCAL)
COM_AddCommand("randomcharacter", randomskin, COM_LOCAL)

local function jrandomskin(p)
	local index = 0
	local bail = 0
	while boringskins[skins[index].name] do
		index = HM_RandomRange(1, #skins) - 1
		bail = $ + 1
		if bail == 100 then
			CONS_Printf(p, "...Huh. Couldn't find a community character to change to. How embarassing.")
			return
		end
	end
	local newskin = skins[index]
	S_StartSound(nil, newskin.soundsid[sfxinfo[sfx_khitem].skinsound], p)
	COM_BufAddText(p, "skin "..newskin.name)
	local sanitizedcolor = SG_Color2Chat[newskin.prefcolor] or ""
	CONS_Printf(p, "Trying to change to "..sanitizedcolor..newskin.realname.." \128("..newskin.kartspeed.."/"..newskin.kartweight..")...")
	ironreminder(p)
end

COM_AddCommand("randomcolor", function(p)
	local index = P_RandomRange(1, #SG_ColorNames)
	S_StartSound(nil, sfx_cdpcm4, p)
	COM_BufAddText(p, ("color %q"):format(SG_ColorNames[index]))
	CONS_Printf(p, "Tried setting your color to "..SG_Color2Chat[index]..SG_ColorNames[index].."\128...")
end)

COM_AddCommand("ironman", function(p)
	if p.HMrandomironman then
		p.HMrandomironman = IRONMAN_OFF
		S_StartSound(nil, sfx_cdfm20, p)
		CONS_Printf(p, "\130IRONMAN \128deactivated...lame.")
	else
		p.HMrandomironman = IRONMAN_ON
		S_StartSound(nil, sfx_cdfm20, p)
		p.HMrandomcount = 69
		CONS_Printf(p, "\130IRONMAN \128activated. You'll automatically randomize every round.")
	end
end)

COM_AddCommand("jironman", function(p)
	if p.HMrandomironman == IRONMAN_JART then
		p.HMrandomironman = false
		S_StartSound(nil, sfx_cdfm20, p)
		CONS_Printf(p, "\130IRONMAN (FORMA JARTE)\128 deactivated...lame.")
	else
		p.HMrandomironman = IRONMAN_JART
		S_StartSound(nil, sfx_cdfm20, p)
		p.HMrandomcount = 69
		CONS_Printf(p, "\130IRONMAN (FORMA JARTE)\128 activated. You'll automatically randomize to a community character every round.")
	end
end)

addHook("ThinkFrame", function()
	HM_Random()
	for p in players.iterate do
		if leveltime == 1 + #p and p.HMrandomironman then
			if p.HMrandomironman == IRONMAN_JART then
				jrandomskin(p)
			else
				randomskin(p)
			end
		end
	end
end)