local hostmod_votetimer = CV_RegisterVar({
	name = "hm_vote_timer",
	defaultvalue = "60",
	flags = CV_NETVAR,
	possiblevalue = CV_Natural
})
local hostmod_autopass = CV_RegisterVar({
	name = "hm_vote_autopass",
	defaultvalue = "0.75",
	flags = CV_NETVAR|CV_FLOAT,
	possiblevalue = { MIN = FRACUNIT/2, MAX = FRACUNIT }
})
local hostmod_speedup = CV_RegisterVar({
	name = "hm_vote_speedup",
	defaultvalue = "0.5",
	flags = CV_NETVAR|CV_FLOAT,
	possiblevalue = { MIN = 0, MAX = FRACUNIT }
})
local hostmod_votecooldown = CV_RegisterVar({
	name = "hm_vote_cooldown",
	defaultvalue = "600",
	flags = CV_NETVAR,
	possiblevalue = CV_Natural
})
local hostmod_allowovertime = CV_RegisterVar({
	name = "hm_vote_allowovertime",
	defaultvalue = "Yes",
	flags = CV_NETVAR,
	possiblevalue = CV_YesNo,
})
local hm_vote_allowidc = CV_RegisterVar({
	name = "hm_vote_allowidc",
	defaultvalue = "Off",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

-- Constants
local VOTE_YES = 1
local VOTE_NO = -1
local VOTE_IDC = 0

local VTYPE_COMMAND = 0
local VTYPE_KICK = 1
local VTYPE_POLL = 2

local vote = {
	timer = 0,
	resolve = 0,
	type = 0,
	votable = {},
	kicktarget = nil,
	cooldown = {},
	command = nil,
}

addHook("NetVars", function(sync) vote = sync($) end)

local inrace = false
local lastvote = 0
local spedup = false

-- We need a whitelist, or a whole bunch of difficult-to-anticipate malicious shit becomes possible.
COM_AddCommand("hm_votable", function(p, ...)
	if not ... then
		CONS_Printf(p, "Supply a command to be whitelisted for voting.")
		return
	end
	local target = table.concat({...}, " "):lower()
	if target:match(" ") then
		CONS_Printf(p, "Supply just the root command, please.")
		return
	end

	-- G: Have to maintain backwards compatibility with 1P voting now...
	if target:sub(1, 1) == "^" then
		target = $:sub(2)
	end

	vote.votable[target] = true

	CONS_Printf(p, "Added \130"..target.."\128 to the votable list.")
end, COM_ADMIN)

-- Remove command from being able to vote
COM_AddCommand("hm_unvotable", function(p, ...)
	if not ... then
		CONS_Printf(p, "Supply a command to be removed from whitelist for voting.")
		return
	end
	local target = table.concat({...}, " "):lower()
	if target:match(" ") then
		CONS_Printf(p, "Supply just the root command, please.")
		return
	end
	vote.votable[target] = nil

	CONS_Printf(p, "Removed \130"..target.."\128 to the votable list.")
end, COM_ADMIN)

local function playercount()
	local a = 0
	for p in players.iterate do
        if not (p.spectator and not (Elimination and Elimination() and p.elim and p.elim.iseliminated)) and p.splitscreenindex == 0 then
		    a = $ + 1
        end
	end
	return a
end

local function totalvote(type)
	local count = 0
	for p in players.iterate do
		if (not p.spectator or vote.type == VTYPE_POLL) and p.splitscreenindex == 0 and p.hostmod.vote == type then count = $ + 1 end
	end
	return count
end

local function votestats()
	if hm_vote_allowidc.value then
		return "("..totalvote(VOTE_YES).." yes, "..totalvote(VOTE_NO).." no, "..(totalvote(VOTE_IDC) + totalvote(nil)).." didn't care)"
	else
		return "("..totalvote(VOTE_YES).." yes, "..totalvote(VOTE_NO).." no)"
	end
end

-- These just pass to the thinkframe to handle vote resolution, they're purely cosmetic otherwise.
local function passvote()
	S_StartSound(nil, sfx_s1a1)
	chatprint("\131*Vote passed! "..votestats())
	vote.resolve = 3 * TICRATE/2
end

local function failvote()
	S_StartSound(nil, sfx_s231)
	chatprint("\131*Vote failed. "..votestats())
end

local function endpoll()
    S_StartSound(nil, sfx_s1a1)
    chatprint("\131*Poll ended! "..votestats())
end

local function istie()
	local yeah = totalvote(VOTE_YES)
	
	local nah = totalvote(VOTE_NO)
	
	if not hm_vote_allowidc.value then nah = nah + totalvote(nil) end
	
	return yeah == nah
end

local function resolvevote()
	if not hm_vote_allowidc.value then
		hm_dprint("the goddamn no-vote conversion")
		for p in players.iterate do
			hm_dprint("inside,", p.name, p.hostmod.vote)
			if p.hostmod.vote == VOTE_IDC or p.hostmod.vote == nil then
				p.hostmod.vote = VOTE_NO
				hm_dprint("did the thing")
			end
		end
	end

	vote.timer = 0

    if vote.type ~= VTYPE_POLL then
        if totalvote(VOTE_YES) > totalvote(VOTE_NO) then
            passvote()
        else
            failvote()
        end
    else
        endpoll()
    end
end

local function castvote(p, choice)
	if not vote.timer then
		chatprintf(p, "\131*There's no vote running...")
		return
	end
    if p.spectator and vote.type ~= VTYPE_POLL then
        chatprintf(p, "\131*Spectators cannot vote...")
        return
    end

	local hm = p.hostmod
	if hm.vote == choice then return end -- same as last choice?

	if hm.vote == nil then
		chatprintf(p, "\131*Your vote was cast.", true)
	end
	hm.vote = choice
	CONS_Printf(server, p.name.." voted "..choice)

	-- keep vanilla client players updated during intermission
	if not (inrace or FEATURE_INTERMISSIONHUD or FEATURE_VOTEHUD) then
		local roundup = vote.timer + TICRATE
		print("\131"..totalvote(VOTE_YES).." yes \133"..totalvote(VOTE_NO).." no \128"..G_TicsToMinutes(roundup)..":"..string.format("%02d", G_TicsToSeconds(roundup)))
	end
	S_StartSound(nil, sfx_s257)
	lastvote = 0

	-- Resolve the vote early in cases of positive landslides or everyone voting.
	if vote.type == VTYPE_POLL then return end

	local count = playercount()
	local autopasslimit = FixedMul(count, hostmod_autopass.value)
	if totalvote(VOTE_YES) > autopasslimit then
		resolvevote()
	elseif totalvote(VOTE_YES) + totalvote(VOTE_NO) == count and not (hostmod_allowovertime.value and istie()) then
		resolvevote()
	end
end

-- Bindable console commands.
COM_AddCommand("voteyes", function(p)
	castvote(p, VOTE_YES)
end)

COM_AddCommand("voteno", function(p)
	castvote(p, VOTE_NO)
end)

COM_AddCommand("voteidc", function(p)
	castvote(p, VOTE_IDC)
end)

COM_AddCommand("autopassneeded", function(p)
	for i = 1, 16 do
		CONS_Printf(p, string.format("%2d %2d", i, FixedMul(i, hostmod_autopass.value) + 1))
	end
end, COM_ADMIN)

local function printwl(p)
	if next(vote.votable) == nil then
		CONS_Printf(p, "This server hasn't enabled any votable commands.")
		return
	end

	local list = "Whitelisted commands are: "
	for command in pairs(vote.votable) do list = $.."\130"..command.."\128, " end
    CONS_Printf(p, list:sub(1, -3)..".")
end

-- Actually handle calling a vote.
local function initvote(command, initiator, poll)
	if not poll then
		-- Prevents whitelist bypasses.
		if command:match(";") then
			CONS_Printf(initiator, "Nice Try!")
			return
		end
		if initiator ~= server and initiator.spectator then
			CONS_Printf(initiator, "Spectators cannot initiate votes...")
			return
		end
	end

	local vt = poll and "poll" or "vote"
	if vote.timer or vote.resolve then
		CONS_Printf(initiator, "A "..vt.." is already in progress. Wait until it's finished!")
		return
	end
	if initiator ~= server and initiator.hostmod.votecooldown > 0 then
		CONS_Printf(initiator, "You just called a "..vt.." recently. Wait a bit! ("..((initiator.hostmod.votecooldown / TICRATE)+1).." seconds left.)")
		return
	end

    vote.command = command:gsub("'", '"') -- Let's at least theoretically let people pass in quotes...
    local spaceindex = string.find(vote.command, " ")
	local commandfrag = vote.command
	if spaceindex then
		commandfrag = vote.command:sub(0, spaceindex - 1)
	end
    commandfrag = commandfrag:lower()

	if not poll and vote.cooldown[commandfrag] then
		CONS_Printf(initiator, "This command has already been voted for recently. Wait a bit! ("..((vote.cooldown[commandfrag] / TICRATE)+1).." seconds left)")
		return
	end

	vote.type = poll and VTYPE_POLL or VTYPE_COMMAND

	-- Find what "base" command is being invoked and check it against the whitelist
	local whitelisted = poll or initiator == server or vote.votable[commandfrag] ~= nil

	if not whitelisted then
		CONS_Printf(initiator, "That command isn't whitelisted for voting.")
		printwl(initiator)
		return
	end

	-- If we're doing kick stuff, handle some specific syntax sugar...
	if not poll and commandfrag == "kick" then
		local target = command:sub(6):lower()
		if target == "" then
			CONS_Printf(initiator, "Who do you want to kick? Type their name or player number.")
			return
		end

		-- We want to mark a player so they can't dodge by name.
		-- We also want them to be targetable by number to avoid griefing with barcode names.
		-- (e.g. "Player 2 renamed to Ii1iLi1iliIl1iililLi1l")
		local playernum = tonumber(target)
		if playernum ~= nil and playernum >= 0 and playernum < #players then
			vote.kicktarget = players[playernum] -- thanks, terminal!
			if not vote.kicktarget then
				CONS_Printf(initiator, "No player matching that number exists.")
				return
			end
		else
			local matches = 0
			for p in players.iterate do
				if p.name:lower():match(target) then
					matches = $ + 1
					vote.kicktarget = p
				end
			end
			if matches == 0 then
				CONS_Printf(initiator, "No player matching that name exists, check your spelling?")
				return
			end
			if matches > 1 then -- Too dangerous to allow ambiguity.
				CONS_Printf(initiator, "That name matches more than one player. Be more specific!")
				return
			end
		end

		if IsPlayerAdmin(vote.kicktarget) or vote.kicktarget == server then
			if hm_debug.value then
				CONS_Printf(initiator, "...I'll let that slide because you're in testing mode.")
			else
				vote.kicktarget = initiator
			end
		end
		vote.command = vote.kicktarget.name
		vote.type = VTYPE_KICK
	end

	vote.timer = hostmod_votetimer.value * TICRATE
	if initiator ~= server and playercount() > 1 then initiator.hostmod.votecooldown = (hostmod_votetimer.value * 2) * TICRATE end
	if not poll then vote.cooldown[commandfrag] = hostmod_votecooldown.value * TICRATE end

	for p in players.iterate do
		p.hostmod.vote = nil
	end

	S_StartSound(nil, sfx_s258)
	local cmd = (vote.type == VTYPE_KICK and "kick " or "")..vote.command
    local vote_or_poll = poll and " called a poll - " or " called a vote to "
	chatprint("\131*"..initiator.name..vote_or_poll.."\130"..cmd.."\131. Vote in chat with \130y \131/ \130n\131.")
	lastvote = 0

	if not hm_debug.value then -- In debug mode, let's allow the vote UI to show up even if there's only one person.
		castvote(initiator, VOTE_YES)
	end
end

COM_AddCommand("vote", function(p, ...)
	if not ... then
		CONS_Printf(p, "What are you trying to vote for?")
		printwl(p)
		return
	end
	local target = table.concat({...}, " ")
	initvote(target, p)
end)

-- Syntax sugar.
COM_AddCommand("votekick", function(p, ...)
	-- Let's not even bother passing garbage to the main handler.
	-- G: actually, we will lol
	COM_BufAddText(p, "vote kick "..table.concat({...}, " "))
end)

-- "What's your question, Soldier?"
COM_AddCommand("votepoll", function(p, ...)
    if not ... then
        CONS_Printf(p, "What's your question?")
        return
    end
    local target = table.concat({...}, " ")
    initvote(target, p, true)
end, COM_ADMIN)

COM_AddCommand("rtv", function(p)
	COM_BufAddText(p, "vote exitlevel")
end)

-- Chat vote handler
addHook("PlayerMsg", function(source, msgtype, _, msg)
	if msgtype ~= 0 then return end

	msg = string.lower(msg)
	if vote.timer then
		if msg == "y" or msg == "rtv" then
			castvote(source, VOTE_YES)
			return true
		elseif msg == "n" then
			castvote(source, VOTE_NO)
			return true
		elseif msg == "idc" or msg == "dc" or msg == "j" then
			castvote(source, VOTE_IDC)
			return true
		end
	elseif msg == "rtv" then
		COM_BufAddText(source, "vote exitlevel")
		return true
	end
end)

COM_AddCommand("voteresetcooldown", function(p)
	vote.cooldown = {}
	for p in players.iterate do
		p.hostmod.votecooldown = 0
	end
	CONS_Printf(p, "Reset all cooldowns")
end, COM_ADMIN)

-- Vote timer and resolution handler.
local function tickVote()
	for cmd, timer in pairs(vote.cooldown) do
		vote.cooldown[cmd] = max(timer - 1, 0)
	end

	for p in players.iterate do
		p.hostmod.votecooldown = max(0, ($ or 0) - 1)
	end

	if vote.resolve > 0 then
		vote.resolve = $ - 1
		if vote.resolve == 0 then
			if vote.type == VTYPE_COMMAND then
				COM_BufAddText(server, vote.command)
			elseif vote.type == VTYPE_KICK then
				if vote.kicktarget and vote.kicktarget.valid then
					COM_BufAddText(server, "kick "..#vote.kicktarget)
				else
					chatprint("\131* ...but the target wasn't here to be kicked!")
				end
			end
		end
	end

	if vote.timer > 0 then
		vote.timer = max($ - 1, 0)
		-- speed up if enough players have voted
		local speeduplimit = FixedMul(playercount(), hostmod_speedup.value)
		if vote.type ~= VTYPE_POLL and (totalvote(VOTE_YES) > speeduplimit or totalvote(VOTE_NO) > speeduplimit) then
			vote.timer = $ - 1
			spedup = true
		else
			spedup = false
		end
		
		if vote.timer <= 0 then
			-- Kinda dirty hack :(
			if hostmod_allowovertime.value == 1 and istie() then
				vote.timer = 1
			else
				resolvevote()
			end
		end
	end
end

-- if the target of a votekick leaves, quickly run the kick command on them so they can't dodge the temp ban

-- no, don't do that. sadly, this doesn't work because through some insane series of events,
-- the server thinks it's a splitscreen player of the person getting kicked,
-- and removes itself from the game, but remains connected to everyone in some kinda limbo state
--[[
addHook("PlayerQuit", function(p)
	if vote.timer > 0 and vote.type == VTYPE_KICK and p == vote.kicktarget then
		vote.timer = 0
		vote.resolve = 0
		S_StartSound(nil, sfx_s1a1)
		chatprint("\131*Vote passed by trial in absentia!")
		COM_BufInsertText(server, "kick "..#p)
	end
end)
--]]

addHook("ThinkFrame", function()
	inrace = true
	tickVote()
end)
addHook("IntermissionThinker", function()
	inrace = false
	tickVote()
end)
addHook("VoteThinker", function()
	inrace = false
	tickVote()
end)

-- HUD
local function drawStringCenter(v, xx, yy, ss, ff, tt)
	local ww = v.stringWidth(ss, ff, tt)
	v.drawString(xx - (ww / 2), yy, ss, ff, tt)
end

local titles = { [VTYPE_COMMAND] = "VOTE", [VTYPE_POLL] = "POLL", [VTYPE_KICK] = "KICK?" }

local function drawVotes(v)
	if not vote.timer then return end

	local flags = V_SNAPTOBOTTOM|V_SNAPTORIGHT|(V_10TRANS * max(0, min((lastvote-TICRATE*5)/(TICRATE/2), 5)))
	lastvote = $ + 1
	local roundup = vote.timer + TICRATE

	local tx = 250 -- Left X
	local ty = 136 -- Top Y
	local bx = 310 -- Right X
	local by = 191 -- Bottom Y

	v.draw(tx, ty, vote.type == VTYPE_POLL and v.cachePatch("POLLBG") or v.cachePatch("VOTEBG"), flags)

	local title = titles[vote.type]
	flags = $ | V_ALLOWLOWERCASE
	v.drawString((bx+tx)/2, ty+4, title, flags, "center") -- title
	drawStringCenter(v, (bx+tx)/2, ty+16, vote.command, flags|V_YELLOWMAP, "small") -- command

	-- kick player status
	if vote.type == VTYPE_KICK then
		if not (vote.kicktarget and vote.kicktarget.valid) then
			drawStringCenter(v, (bx+tx)/2, ty+20, "Left the game", flags, "small")
		elseif vote.command ~= vote.kicktarget.name then
			drawStringCenter(v, (bx+tx)/2, ty+20, "AKA "..vote.kicktarget.name, flags, "small")
		end
	end

	v.drawString(tx+4, ty+24, "YEAH", flags|V_GREENMAP, "small") -- yeah
	v.drawString(bx-4, ty+24, "NAH", flags|V_REDMAP, "small-right") -- nah
	v.drawString(tx+4, ty+29, totalvote(VOTE_YES), flags|V_GREENMAP) -- yes votes
	v.drawString(bx-4, ty+29, totalvote(VOTE_NO), flags|V_REDMAP, "right") -- no votes

	drawStringCenter(v, (bx+tx)/2, ty+39, "chat \131y\128 or \133n", flags + V_40TRANS, "small") -- instructions

	-- speedup flashing
	if spedup and vote.timer & 16 then
		flags = $ | V_GRAYMAP
	end
	
	if vote.timer == 1 and hostmod_allowovertime.value and istie() then
		v.drawString(tx + (bx - tx)/2 - v.stringWidth("OVERTIME", flags, "small")/2, by-9, "OVERTIME", flags, "small")
	else
		local secstr = string.format("%02d", G_TicsToSeconds(roundup))
		local str = string.format("%02d", G_TicsToMinutes(roundup))..":"..secstr
		v.drawString(tx + (bx - tx)/2 + v.stringWidth(":"..secstr, flags, "small"), by-9, str, flags, "small-right") -- timer
	end
end

hud.add(function(v, p) if p == displayplayers[splitscreen] then drawVotes(v) end end, "game")
hud.add(drawVotes, "scores")

-- Hehe
if FEATURE_INTERMISSIONHUD then hud.add(drawVotes, "intermission") end
if FEATURE_VOTEHUD then hud.add(drawVotes, "vote") end
